#version 330 core

// Parts of the shader code taken from: http://en.wikibooks.org/wiki/OpenGL_Programming

layout(location=0) out vec4 outColor;

struct material
{
	vec4 ambientColor;
	vec4 diffuseColor;
	vec4 specularColor;
	float shininess;
};

struct light
{
	vec4 position;
	vec4 ambientColor;
	vec4 diffuseColor;
	vec4 specularColor;
	vec3 spotDirection;
	float constantAttenuation;
	float linearAttenuation;
	float quadraticAttenuation;
	float spotCutoff;
	float spotExponent;
};

uniform mat4 view;
uniform mat4 projection;
uniform mat4 view_inverse;
uniform int numLights;
uniform int drawWithoutTextureWithoutLight;

in vec3 worldNormal;
in vec2 texCoord;
in vec4 worldPosition;
in mat3 tbn;
uniform sampler2D color_texture;
uniform sampler2D normal_texture;

uniform int normalMapping;	// Whether we use normal mapping
uniform int hdr;			// Whether we use HDR

in vec4 lightSpacePosition[8];
uniform light lights[8];
uniform sampler2DShadow shadowMap[8];
uniform material material0;

float calcShadowFactor(int lightIndex)
{
	float shadowMult = textureProj(shadowMap[lightIndex], lightSpacePosition[lightIndex]);
	return shadowMult;
} 

void main()
{
	vec4 finalColor;

	if (drawWithoutTextureWithoutLight != 1)
	{
		vec4 textureColor = texture2D(color_texture, texCoord).rgba;
		vec3 normalDirection = normalize(worldNormal);
		vec3 normalDirectionTangent = normalize(texture2D(normal_texture, texCoord).rgb*2.0 - 1.0);
		vec3 viewDirection = normalize(vec3(view_inverse * vec4(0.0, 0.0, 0.0, 1.0) - worldPosition));
		//vec3 viewDirectionTangent = tbn * viewDirection;
		vec3 lightDirection;
		vec3 lightDirectionTangent;
		
		float attenuation;
		
		vec3 maxAmbientLight = vec3(0.0, 0.0, 0.0);
		vec3 diffuseLighting = vec3(0.0, 0.0, 0.0);
		vec3 specularLighting = vec3(0.0, 0.0, 0.0);
		
		if (textureColor.a > 0.01)
		{
			// Self-Illuminating
			diffuseLighting = vec3(1.0, 1.0, 1.0);
		}
		else
		{
			for (int i = 0; i < numLights; i++)
			{
				float shadowFactor = calcShadowFactor(i);
				
				if (0.01 > lights[i].position.w) // directional light?
				{
					attenuation = 1.0; // no attenuation
					lightDirection = normalize(vec3(lights[i].position));
					lightDirectionTangent = tbn*lightDirection;
					vec3 ambientLight = vec3(lights[i].ambientColor) * vec3(material0.ambientColor);
					maxAmbientLight = max(ambientLight, maxAmbientLight);
					shadowFactor = 1.0f; // No Shadow
				} 
				else // point light or spotlight (or other kind of light) 
				{
					vec3 positionToLightSource = vec3(lights[i].position - worldPosition);
					float distance = length(positionToLightSource);
					lightDirection = normalize(positionToLightSource);
					lightDirectionTangent = tbn*lightDirection;
					attenuation = 1.0 / (lights[i].constantAttenuation
									   + lights[i].linearAttenuation * distance
									   + lights[i].quadraticAttenuation * distance * distance);
					
					if (lights[i].spotCutoff <= 90.0) // spotlight?
					{
						float clampedCosine = max(0.0, dot(-lightDirection, lights[i].spotDirection));
						if (clampedCosine < cos(radians(lights[i].spotCutoff))) // outside of spotlight cone?
						{
							attenuation = 0.0;
						}
						else
						{
							attenuation = attenuation * pow(clampedCosine, lights[i].spotExponent);   
						}
					}
					
				}
				
				vec3 diffuseReflection;
				if (normalMapping != 0)
				{
					diffuseReflection = attenuation 
						* vec3(lights[i].diffuseColor) * vec3(material0.diffuseColor)
						* max(0.0, dot(normalDirectionTangent, lightDirectionTangent));
				}
				else
				{
					diffuseReflection = attenuation 
						* vec3(lights[i].diffuseColor) * vec3(material0.diffuseColor)
						* max(0.0, dot(normalDirection, lightDirection));
				}
			 
				vec3 specularReflection;
				if (dot(normalDirection, lightDirection) < 0.0) // light source on the wrong side?
				{
					specularReflection = vec3(0.0, 0.0, 0.0); // no specular reflection
				}
				else // light source on the right side
				{
					specularReflection = attenuation * vec3(lights[i].specularColor) * vec3(material0.specularColor) 
						* pow(max(0.0, dot(reflect(-lightDirection, normalDirection), viewDirection)), material0.shininess);
				}
				diffuseLighting = diffuseLighting + shadowFactor*diffuseReflection;
				specularLighting = specularLighting + shadowFactor*specularReflection;
				//lighting = lighting + shadowFactor*(diffuseReflection + specularReflection);
			}
		}
		
		if (numLights == 0)
			diffuseLighting = vec3(1.0, 1.0, 1.0);
		
		vec3 light;
		if (hdr == 0)
			light = clamp(maxAmbientLight+diffuseLighting, 0.0, 1.0) + specularLighting;
		else
			light = maxAmbientLight+diffuseLighting + specularLighting;

		finalColor = vec4(light, 1.0) * textureColor;
	}
	else
		finalColor = vec4(1.0,1.0,1.0,1.0);
		
	outColor = finalColor;
}